#include "mapping_info.h"

#include "design.h"
#include "hierarchical_channel.h"
#include "inoutport.h"
#include "inport.h"
#include "module_instance.h"
#include "module.h"
#include "outport.h"
#include "port_mapping.h"
#include "../tools/function_helper.h"
/*TLM 2.0*/
#include "initiatorsocket.h"
#include "targetsocket.h"
#include "socket.h"
/*end TLM 2.0*/



risc::MappingInfo::MappingInfo():
    is_initialized_(false)
{ }

void risc::MappingInfo::initialize(
    PathInstanceMapper *path_instance_mapper)
{
  //std::cout << "initialize portmapping" << std::endl;
  if(is_initialized_) {
    return;
  }

  if(PSG_MODE){
    is_initialized_=true;
    return;
  }
  
  Design *d = risc::get_design();

  if(!d->perform_static_analysis()) {
    return;
  }
  

  /*support for indirect mapping*/
  //forward mapping
  //most common
  std::unordered_map<std::pair<int, SgVariableDefinition*>, risc::ModuleInstance*, pairhash> socket_MI_mapping;
  //for all the module
  for(risc::ModuleVectorConstIter
       iter_module  = d->module_definitions_.begin();
       iter_module != d->module_definitions_.end();
       iter_module++) 
  {
    //for all the module instance 
    for(std::list<Instance*>::iterator
      inst_iter  = (*iter_module)->instances_.begin();
      inst_iter != (*iter_module)->instances_.end();
      inst_iter ++)
    {
      std::list<risc::InstanceTree> instance_trees                                   
        = d->get_root().find_all_instances_of(
          risc::tools::get_sym_from_def((*inst_iter)->get_ast_node()));

      //for all the instance trees of the module instance
      for(std::list<risc::InstanceTree>::iterator
        tree_iter = instance_trees.begin();
        tree_iter != instance_trees.end();
        tree_iter ++)
      {
        //for all the path mappings in this module
        for(risc::Class::PathMappingType::iterator
          pm_iter = (*iter_module)->path_mappings.begin();
          pm_iter != (*iter_module)->path_mappings.end();
          pm_iter ++)
        {
          risc::InstanceTree t1 = (*tree_iter).extend_with_symbol_list((*pm_iter).first);
          risc::InstanceTree t2 = (*tree_iter).extend_with_symbol_list((*pm_iter).second);
          
          if(t1.path_.empty() || t2.path_.empty()) continue;
          //the forward mapping
          //store the t1 to t2 mapping into the map inst_tree_mapping
          //prepare the key
          int obj_id_l = path_instance_mapper->get_id(t1);
          SgVariableDefinition *var_l = (t1.get_object())->get_ast_node();
          std::pair<int, SgVariableDefinition*> p1 = {obj_id_l, var_l};

          //different situations
          risc::Socket* s1 = dynamic_cast<risc::Socket*>(t1.get_object());
          risc::Socket* s2 = dynamic_cast<risc::Socket*>(t2.get_object());
          //direct target to target mapping
          if(s1!=NULL && s1->type_ == risc::Definition::TargetSocketType && 
             s2!=NULL && s2->type_ == risc::Definition::TargetSocketType)
          {
            if(inst_tree_mapping.count(p1)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              inst_tree_mapping[p1] = t2;
            }
          }
          //target to initiator connection. I dont think we have such test case yet
          else if(s1!=NULL && s1->type_ == risc::Definition::TargetSocketType){
            //prepare key
            int obj_id_r = path_instance_mapper->get_id(t2);
            SgVariableDefinition *var_r = (t2.get_object())->get_ast_node();
            std::pair<int, SgVariableDefinition*> p2 = {obj_id_r, var_r};

            if(inst_tree_mapping.count(p2)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              inst_tree_mapping[p2] = t1;
            }
          }
          //init to init 
          else if(s1!=NULL && s1->type_ == risc::Definition::InitiatorSocketType  && 
             s2!=NULL && s2->type_ == risc::Definition::InitiatorSocketType)
          {
            if(inst_tree_mapping.count(p1)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              inst_tree_mapping[p1] = t2;
            }
          }
          //init to tar
          else if(s1!=NULL && s1->type_ == risc::Definition::InitiatorSocketType  && 
             s2!=NULL && s2->type_ == risc::Definition::TargetSocketType)
          {
            if(inst_tree_mapping.count(p1)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              inst_tree_mapping[p1] = t2;
            }
          }
          else
          {
            //adding for ports:

            // t1.print_instance_tree();
            // t2.print_instance_tree();
            inst_tree_mapping[p1] = t2;
          }

          //prepare socket_MI_mapping
          //for example, we have two socket mappings
          //init to wrapper_init
          //and wrapper_init to tar
          //then we need to know the module instance of init
          //and in later steps, we can add socket mapping of
          //init to tar in that module instance
          if(s1!=NULL && s1->type_ == risc::Definition::InitiatorSocketType){
            t1.pop_back();

            socket_MI_mapping[p1] = dynamic_cast<risc::ModuleInstance*>(t1.get_object());
          }
          //
          
          /*
          int socket_id
            = path_instance_mapper->get_id(t1);
          int mapped_socket_id 
            = path_instance_mapper->get_id(t2);
          SgVariableDefinition *socket_var = (t1.get_object())->get_ast_node();  
          mapping_[{socket_id,socket_var}]
            = std::pair<risc::InstanceTree, int>
              (t2, mapped_socket_id);
          std::cout << "mapping added "
                    << "[" << socket_id
                    << "," << (t1.get_object())->get_name()
                    << "," << std::endl;
          t2.print_instance_tree();
          std::cout << "," << mapped_socket_id << " ]" << std::endl;
          */
        }
      }
    }
  }

  for(risc::MappingInfo::TreeMap::iterator 
    mapping_iter = inst_tree_mapping.begin();
    mapping_iter != inst_tree_mapping.end();
    mapping_iter ++)
  {
    std::pair<int, SgVariableDefinition*> p1 = mapping_iter->first;
    risc::InstanceTree t2 = mapping_iter->second;
    SgVariableDefinition *var1 = p1.second;

    if(!var1)
    {
      std::cout << "var1 is NULL, should not happen" << std::endl;
    }
    //meaning it is a port
    else if(!d->find_socket(var1)){
      /*std::cout << "t1 is path for port " 
        << var1->get_declaration()->get_decl_item(0)->get_qualified_name().getString()
        << std::endl;*/
      SgVariableDefinition *var2 = (t2.get_object())->get_ast_node();
      //we want to make mapping of port-to-channel
      bool flag = false;
      while(t2.get_scope()->find_port(var2)){
        int obj_id_l = path_instance_mapper->get_id(t2);
        SgVariableDefinition *var_l = var2;
        std::pair<int, SgVariableDefinition*> p2 = {obj_id_l, var_l};
        //
        if(inst_tree_mapping.count(p2)==0)
        {
          /*std::cout << "instance tree t2 = ";
          t2.print_instance_tree();
          std::cout << "not bound" << std::endl;*/
          flag = true;
          break;
        }
        t2 = inst_tree_mapping[p2];
        var2 = (t2.get_object())->get_ast_node();
      }
      if(flag) continue;
      //t2 now is a inst tree for a channel
      int port_id = p1.first;
      int mapped_port_id 
        = path_instance_mapper->get_id(t2);
      mapping_[std::make_pair(port_id,risc::tools::get_sym_from_def(var1))]
        = std::pair<risc::InstanceTree, int>
          (t2, mapped_port_id);
      /*std::cout << "mapping added "
                << "[" << port_id
                << "," << SageInterface::get_name(var1)
                << "," << std::endl;
      t2.print_instance_tree();
      std::cout << "," << mapped_port_id << " ]" << std::endl;*/
    }
    else 
    {
      //std::cout << "t1 is path for socket" << std::endl;
      risc::Socket* s1 = d->find_socket(var1);
      //for solvine DM's target to target mapping missing bug
      //std::cout << "s1 is " << s1->get_name() << std::endl;
      //if(s1->type_ == risc::Definition::TargetSocketType) continue;
      risc::Socket* s2 = dynamic_cast<risc::Socket*>(t2.get_object());
      int socket_id = p1.first;
      bool flag = false;
      while(true){
        if(s2 == NULL) {
          std::cout << "Error socket mapping" << std::endl;
          exit(0);
        }
        if(s2->type_ == risc::Definition::InitiatorSocketType){
          int obj_id_l = path_instance_mapper->get_id(t2);
          SgVariableDefinition *var_l = s2->get_ast_node();
          std::pair<int, SgVariableDefinition*> p2 = {obj_id_l, var_l};
          //
          if(inst_tree_mapping.count(p2)==0)
          {
            /*std::cout << "instance tree t2 = ";
            t2.print_instance_tree();
            std::cout << "not bound" << std::endl;*/
            flag = true;
            break;
          }
          t2 = inst_tree_mapping[p2];
          s2 = dynamic_cast<risc::Socket*>(t2.get_object());
        }
        else{
          //is target type
          int obj_id_l = path_instance_mapper->get_id(t2);
          SgVariableDefinition *var_l = s2->get_ast_node();
          std::pair<int, SgVariableDefinition*> p2 = {obj_id_l, var_l};
          //
          if(inst_tree_mapping.count(p2)==0)
          {
            /*std::cout << "instance tree t2 = ";
            t2.print_instance_tree();
            std::cout << "not bound" << std::endl;*/
            
            break;
          }
          t2 = inst_tree_mapping[p2];
          s2 = dynamic_cast<risc::Socket*>(t2.get_object());
        }
      }
      if(flag) continue;
      int mapped_socket_id 
        = path_instance_mapper->get_id(t2);
      SgVariableDefinition *socket_var = var1;
      mapping_[std::make_pair(socket_id,risc::tools::get_sym_from_def(socket_var))]
        = std::pair<risc::InstanceTree, int>
          (t2, mapped_socket_id);
      /*std::cout << "\nmapping added "
                << "[" << socket_id
                << "," << SageInterface::get_name(var1)
                << "," << std::endl;
      t2.print_instance_tree();
      std::cout << "," << mapped_socket_id << " ]" << std::endl;*/
      if(socket_MI_mapping[p1])
        socket_MI_mapping[p1]->add_socket_mapping(s1,s2);
      /*
      std::cout << "updated mapping of "
                    << s1->get_name() << " -> " 
                    << s2->get_name() << " to module instance "
                    << socket_MI_mapping[p1]->get_name()
                    << std::endl;*/
    }
    
  }

  //backward mapping
  //used for nb_transport_bw
  std::unordered_map<std::pair<int, SgVariableDefinition*>, risc::ModuleInstance*, pairhash> backward_socket_MI_mapping;
  //for all the module
  for(risc::ModuleVectorConstIter
       iter_module  = d->module_definitions_.begin();
       iter_module != d->module_definitions_.end();
       iter_module++) 
  {
    //for all the module instance 
    for(std::list<Instance*>::iterator
      inst_iter  = (*iter_module)->instances_.begin();
      inst_iter != (*iter_module)->instances_.end();
      inst_iter ++)
    {
      std::list<risc::InstanceTree> instance_trees                                   
        = d->get_root().find_all_instances_of(
          risc::tools::get_sym_from_def((*inst_iter)->get_ast_node()));

      //for all the instance trees of the module instance
      for(std::list<risc::InstanceTree>::iterator
        tree_iter = instance_trees.begin();
        tree_iter != instance_trees.end();
        tree_iter ++)
      {
        //for all the path mappings in this module
        for(risc::Class::PathMappingType::iterator
          pm_iter = (*iter_module)->path_mappings.begin();
          pm_iter != (*iter_module)->path_mappings.end();
          pm_iter ++)
        {

          //main difference with forward is here,
          //t1 and t2 are exchanged
          //std::cout << "in backward_socket_MI_mapping" << std::endl;
          risc::InstanceTree t1 = (*tree_iter).extend_with_symbol_list((*pm_iter).second);
          risc::InstanceTree t2 = (*tree_iter).extend_with_symbol_list((*pm_iter).first);
          // t1.print_instance_tree();
          // t2.print_instance_tree();
          if(t1.path_.empty() || t2.path_.empty()) continue;

          //the backward mapping
          //store the t1 to t2 mapping into the map inst_tree_mapping
          //prepare the key
          int obj_id_l = path_instance_mapper->get_id(t1);
          SgVariableDefinition *var_l = (t1.get_object())->get_ast_node();
          std::pair<int, SgVariableDefinition*> p1 = {obj_id_l, var_l};

          //different situations
          risc::Socket* s1 = dynamic_cast<risc::Socket*>(t1.get_object());
          risc::Socket* s2 = dynamic_cast<risc::Socket*>(t2.get_object());
          //direct target to target mapping

          //another difference
          //TargetSocketType changed to InitiatorSocketType
          if(s1!=NULL && s1->type_ == risc::Definition::InitiatorSocketType && 
             s2!=NULL && s2->type_ == risc::Definition::InitiatorSocketType)
          {
            if(backward_inst_tree_mapping.count(p1)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              backward_inst_tree_mapping[p1] = t2;
            }
          }
          //target to initiator connection. I dont think we have such test case yet
          else if(s1!=NULL && s1->type_ == risc::Definition::InitiatorSocketType){
            //prepare key
            int obj_id_r = path_instance_mapper->get_id(t2);
            SgVariableDefinition *var_r = (t2.get_object())->get_ast_node();
            std::pair<int, SgVariableDefinition*> p2 = {obj_id_r, var_r};

            if(backward_inst_tree_mapping.count(p2)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              backward_inst_tree_mapping[p2] = t1;
            }
          }
          //tar to tar
          else if(s1!=NULL && s1->type_ == risc::Definition::TargetSocketType && 
             s2!=NULL && s2->type_ == risc::Definition::TargetSocketType) 
          {
            if(backward_inst_tree_mapping.count(p1)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              backward_inst_tree_mapping[p1] = t2;
            }
          }
          //init to tar
          else if(s1!=NULL && s1->type_ == risc::Definition::InitiatorSocketType && 
             s2!=NULL && s2->type_ == risc::Definition::TargetSocketType) 
          {
            if(backward_inst_tree_mapping.count(p1)!=0){
              std::cout << "WARNING! instance tree pair <t1 t2> already added" << std::endl;
              //exit(0);
            }else{
              backward_inst_tree_mapping[p1] = t2;
            }
          }

          //prepare socket_MI_mapping
          //for example, we have two socket mappings
          //init to wrapper_init
          //and wrapper_init to tar
          //then we need to know the module instance of init
          //and in later steps, we can add socket mapping of
          //init to tar in that module instance
          if(s1!=NULL && s1->type_ == risc::Definition::InitiatorSocketType){
            t1.pop_back();

            backward_socket_MI_mapping[p1] = dynamic_cast<risc::ModuleInstance*>(t1.get_object());
          }
          //
          
          /*
          int socket_id
            = path_instance_mapper->get_id(t1);
          int mapped_socket_id 
            = path_instance_mapper->get_id(t2);
          SgVariableDefinition *socket_var = (t1.get_object())->get_ast_node();  
          mapping_[{socket_id,socket_var}]
            = std::pair<risc::InstanceTree, int>
              (t2, mapped_socket_id);
          std::cout << "mapping added "
                    << "[" << socket_id
                    << "," << (t1.get_object())->get_name()
                    << "," << std::endl;
          t2.print_instance_tree();
          std::cout << "," << mapped_socket_id << " ]" << std::endl;
          */
        }
      }
    }
  }

  for(risc::MappingInfo::TreeMap::iterator 
    mapping_iter = backward_inst_tree_mapping.begin();
    mapping_iter != backward_inst_tree_mapping.end();
    mapping_iter ++)
  {
    //std::cout << "in the backward loop" << std::endl;
    std::pair<int, SgVariableDefinition*> p1 = mapping_iter->first;
    risc::InstanceTree t2 = mapping_iter->second;
    SgVariableDefinition *var1 = p1.second;

    if(!var1)
    {
      std::cout << "var1 is NULL, should not happen" << std::endl;
    }
    //meaning it is a port
    else if(!d->find_socket(var1)){
      //in backward mapping we dont need to consider ports
    }
    else //meaning it is a socket
    {
      risc::Socket* s1 = d->find_socket(var1);
      risc::Socket* s2 = dynamic_cast<risc::Socket*>(t2.get_object());
      
      int socket_id = p1.first;
      bool flag = false;
      while(true){
        if(s2 == NULL) {
          std::cout << "Error socket mapping" << std::endl;
          exit(0);
        }
        if(s2->type_ == risc::Definition::TargetSocketType){
          int obj_id_l = path_instance_mapper->get_id(t2);
          SgVariableDefinition *var_l = s2->get_ast_node();
          std::pair<int, SgVariableDefinition*> p2 = {obj_id_l, var_l};
          //
          if(backward_inst_tree_mapping.count(p2)==0)
          {
            /*std::cout << "instance tree t2 = ";
            t2.print_instance_tree();
            std::cout << "not bound" << std::endl;*/
            flag = true;
            break;
          }
          t2 = backward_inst_tree_mapping[p2];
          s2 = dynamic_cast<risc::Socket*>(t2.get_object());
        }
        else{
          //s2 is init type
          int obj_id_l = path_instance_mapper->get_id(t2);
          SgVariableDefinition *var_l = s2->get_ast_node();
          std::pair<int, SgVariableDefinition*> p2 = {obj_id_l, var_l};
          //
          if(backward_inst_tree_mapping.count(p2)==0)
          {
            /*std::cout << "instance tree t2 = ";
            t2.print_instance_tree();
            std::cout << "not bound" << std::endl;*/
            
            break;
          }
          t2 = backward_inst_tree_mapping[p2];
          s2 = dynamic_cast<risc::Socket*>(t2.get_object());
        }
      }
      if(flag) continue;
      int mapped_socket_id 
        = path_instance_mapper->get_id(t2);
      SgVariableDefinition *socket_var = var1;
      mapping_[std::make_pair(socket_id,risc::tools::get_sym_from_def(socket_var))]
        = std::pair<risc::InstanceTree, int>
          (t2, mapped_socket_id);

      // std::cout << "\nmapping added "
      //           << "[" << socket_id
      //           << "," << SageInterface::get_name(var1)
      //           << "," << std::endl;
      // t2.print_instance_tree();
      // std::cout << "," << mapped_socket_id << " ]" << std::endl;

      if(backward_socket_MI_mapping[p1])
        backward_socket_MI_mapping[p1]->add_backward_socket_mapping(s1,s2);
      /*
      std::cout << "updated mapping of "
                    << s1->get_name() << " -> " 
                    << s2->get_name() << " to module instance "
                    << socket_MI_mapping[p1]->get_name()
                    << std::endl;*/
    }
    
  }
  //end
  /*std::cout << "\n\n\n\n\n";

  for(risc::ModuleVectorConstIter
       iter_module  = d->module_definitions_.begin();
       iter_module != d->module_definitions_.end();
       iter_module++) 
  {
    std::cout << "in module " << (*iter_module)->get_name() << std::endl;
    for(risc::InitiatorSocketVectorIter init_iter = (*iter_module)->initiatorsockets_.begin();
      init_iter != (*iter_module)->initiatorsockets_.end();
      init_iter++) 
    {
      std::cout << "init socket is " << (*init_iter)->get_name() << std::endl;
      SgVariableSymbol* target_of_initiator_symbol = mapping_[0][(*init_iter)->get_ast_node()].first.get_object_symbol();
      std::cout << "target_of_initiator_symbol is " << target_of_initiator_symbol->get_name() << std::endl;

    }
  }*/

  std::unordered_map<int,
                     std::unordered_map<SgVariableDefinition*,
                                        std::pair<risc::InstanceTree,
                                                  int> > >::iterator cache_iter;

  for(risc::ModuleVectorConstIter
       iter_module  = d->module_definitions_.begin();
       iter_module != d->module_definitions_.end();
       iter_module++) 
  {
    // std::cout << "analyzing module: " << (*iter_module)->get_name() << std::endl;
    // Strategy
    // First,  we get all instances of the port
    // Second, we determine the instance id of the port hosting module
    // Third,  we determine the mapped channel path
    // Forth,  we determine the instance id of the mapped channel
    // Sixth,  we add the pair of channel path
    //         and channel instance id to the table

    #define ANALYZE_PORT( PORT_TYPE, VAR_NAME )                           \
    for(risc:: PORT_TYPE                                                  \
         iter_port  = (*iter_module)-> VAR_NAME .begin();                 \
         iter_port != (*iter_module)-> VAR_NAME .end();                   \
         iter_port++) {                                                   \
      SgVariableDefinition *port_var = (*iter_port)->get_ast_node();      \
      std::list<risc::InstanceTree> ports                                 \
        = d->get_root().find_all_instances_of(                            \
          risc::tools::get_sym_from_def(port_var));                       \
      for(std::list<risc::InstanceTree>::iterator                         \
          port_path_iter  = ports.begin();                                \
          port_path_iter != ports.end();                                  \
          port_path_iter++) {                                             \
        int port_id                                                       \
          = path_instance_mapper->get_id(*port_path_iter);                \
        if(mapping_.count(std::make_pair                                  \
          (port_id,risc::tools::get_sym_from_def(port_var))))             \
            continue;                                                     \
        InstanceTree channel_path                                         \
          = risc::get_mapped_channel_instance(*port_path_iter);           \
        int channel_id                                                    \
          = path_instance_mapper->get_id(channel_path);                   \
        mapping_[std::make_pair(                                          \
          port_id,risc::tools::get_sym_from_def(port_var))]               \
          = std::pair<risc::InstanceTree, int>(channel_path, channel_id); \
      }                                                                   \
    }

    ANALYZE_PORT( InPortVectorIter , inports_ );
    ANALYZE_PORT( InOutPortVectorIter , inoutports_ );
    ANALYZE_PORT( OutPortVectorIter , outports_ );
    #undef ANALYZE_PORT

    
    /*

    #define ANALYZE_SOCKET( SOCKET_TYPE, VAR_NAME )                           \
    for(risc:: SOCKET_TYPE                                                    \
         iter_socket  = (*iter_module)-> VAR_NAME .begin();                   \
         iter_socket != (*iter_module)-> VAR_NAME .end();                     \
         iter_socket++) {                                                     \
      SgVariableDefinition *socket_var = (*iter_socket)->get_ast_node();      \
      std::list<risc::InstanceTree> sockets                                   \
        = d->get_root().find_all_instances_of(socket_var);                    \
      for(std::list<risc::InstanceTree>::iterator                             \
          socket_path_iter  = sockets.begin();                                \
          socket_path_iter != sockets.end();                                  \
          socket_path_iter++) {                                               \
        int socket_id                                                         \
          = path_instance_mapper->get_id(*socket_path_iter);                  \
        if(mapping_.count(socket_id))                                         \
          if(mapping_[socket_id].count(socket_var))                           \
            continue;                                                         \
        InstanceTree mapped_socket_path                                       \
          = risc::get_mapped_socket_instance_tree(*socket_path_iter);         \
        if(mapped_socket_path.get_scope() == NULL){                           \
          continue;                                                           \
        }                                                                     \
        int mapped_socket_id                                                  \
          = path_instance_mapper->get_id(mapped_socket_path);                 \
        mapping_[socket_id][socket_var]                                       \
          = std::pair<risc::InstanceTree, int>                                \
            (mapped_socket_path, mapped_socket_id);                           \
      }                                                                       \
    }

    for(risc:: InitiatorSocketVectorIter                                                    
         iter_socket  = (*iter_module)-> initiatorsockets_ .begin();                   
         iter_socket != (*iter_module)-> initiatorsockets_ .end();                     
         iter_socket++) {                                                     
      SgVariableDefinition *socket_var = (*iter_socket)->get_ast_node();      
      std::list<risc::InstanceTree> sockets                                   
        = d->get_root().find_all_instances_of(socket_var);                    
      for(std::list<risc::InstanceTree>::iterator                             
          socket_path_iter  = sockets.begin();                                
          socket_path_iter != sockets.end();                                  
          socket_path_iter++) {                                               
        int socket_id                                                         
          = path_instance_mapper->get_id(*socket_path_iter);                  
        std::cout << "socket_id = " << socket_id << std::endl;                
        (*socket_path_iter).print_instance_tree();                            
        InstanceTree mapped_socket_path                                       
          = risc::get_mapped_socket_instance_tree(*socket_path_iter);         
        if(mapped_socket_path.get_scope() == NULL){                           
          continue;                                                           
        }                                                                     
        int mapped_socket_id                                                  
          = path_instance_mapper->get_id(mapped_socket_path);                 
        std::cout << "mapped_socket_id = " << mapped_socket_id << std::endl;  
        mapping_[socket_id][socket_var]                                       
          = std::pair<risc::InstanceTree, int>                                
            (mapped_socket_path, mapped_socket_id);                           
      }                                                                       
    }

    //ANALYZE_SOCKET( InitiatorSocketVectorIter , initiatorsockets_ );
    //ANALYZE_SOCKET( TargetSocketVectorIter , targetsockets_ );
    #undef ANALYZE_SOCKET
    */

  }

  for(risc::HierarchicalChannelVectorConstIter
       iter_hier_channel  = d->hierarchical_channel_.begin();
       iter_hier_channel != d->hierarchical_channel_.end();
       iter_hier_channel++) {

    #define ANALYZE_PORT( PORT_TYPE, VAR_NAME )                           \
    for(risc:: PORT_TYPE                                                  \
         iter_port  = (*iter_hier_channel)-> VAR_NAME .begin();           \
         iter_port != (*iter_hier_channel)-> VAR_NAME .end();             \
         iter_port++) {                                                   \
      SgVariableDefinition *port_var = (*iter_port)->get_ast_node();      \
      std::list<risc::InstanceTree> ports                                 \
        = d->get_root().find_all_instances_of(                            \
          risc::tools::get_sym_from_def(port_var));                       \
      for(std::list<risc::InstanceTree>::iterator                         \
          port_path_iter  = ports.begin();                                \
          port_path_iter != ports.end();                                  \
          port_path_iter++) {                                             \
        int port_id                                                       \
          = path_instance_mapper->get_id(*port_path_iter);                \
        if(mapping_.count(std::make_pair(                                 \
          port_id,risc::tools::get_sym_from_def(port_var))))              \
            continue;                                                     \
        InstanceTree channel_path                                         \
          = risc::get_mapped_channel_instance(*port_path_iter);           \
        int channel_id                                                    \
          = path_instance_mapper->get_id(channel_path);                   \
        mapping_[std::make_pair(                                          \
          port_id,risc::tools::get_sym_from_def(port_var))]               \
          = std::pair<risc::InstanceTree, int>(channel_path, channel_id); \
      }                                                                   \
    }
    ANALYZE_PORT( InPortVectorIter , inports_ );
    ANALYZE_PORT( InOutPortVectorIter , inoutports_ );
    ANALYZE_PORT( OutPortVectorIter , outports_ );
    #undef ANALYZE_PORT
  }


  //why false? ZC 09/oct/2018
  is_initialized_ = true;
}


void risc::MappingInfo::print()
{
  Mapping_::iterator iter;
  std::cout << "mapping_ size = " << mapping_.size() << std::endl;
  for(iter  = mapping_.begin();
      iter != mapping_.end();
      iter++) {

      if(!iter->first.second) continue;
      std::cout << iter->first.first << " | "
                << isSgVariableSymbol(iter->first.second)->get_declaration()->get_qualified_name() << ":"
                << isSgVariableSymbol(iter->first.second)->get_declaration()->get_file_info()->get_line()
                << " | ";
      iter->second.first.print_instance_tree();
      std::cout << ":" << iter->second.second
                << std::endl;
    
  }
}

std::list<std::pair<int, risc::InstanceTree>>
risc::MappingInfo::get_mapped_module(
    int channel_id,
    SgClassDefinition *channel_definition)
{
  std::list<std::pair<int, risc::InstanceTree>> result;
  for(Mapping_::iterator
      iter  = mapping_.begin();
      iter != mapping_.end();
      iter++) {

      if(iter->second.second == channel_id) {

        SgClassType *path_channel_type
          = isSgClassType(iter->second.first.get_object_symbol()->get_type());

        SgClassDeclaration *path_channel_decl
          = isSgClassDeclaration(path_channel_type->get_declaration());

        path_channel_decl
          = isSgClassDeclaration(path_channel_decl->get_definingDeclaration());


        if(path_channel_decl->get_definition() == channel_definition) {

          std::list<InstanceTree> temp2
            = risc::get_design()->get_root()
                                    .find_all_instances_of(iter->first.second);

          InstanceTree path
            = risc::get_design()->path_instance_mapper_
                                    ->get_instance_tree(iter->first.first, temp2);

          path.pop_back();

          std::pair<int, risc::InstanceTree> temp;
          temp.first = iter->first.first;
          temp.second = path;

          result.push_back(temp);
        
      }
    }
  }

  return result;
}

