#ifndef OBJECT_H_INCLUDED_
#define OBJECT_H_INCLUDED_

#include "rose.h"

#include "definition.h"

namespace risc {

class Object: public Definition {

public:

  /**
   * \brief default constructor
   */
  explicit Object(SgVariableDefinition *ast_node, Type type);

  /**
   * \brief  Returns the ast definition of the object
   * \return Pointer to the definition
   */
  SgVariableDefinition *get_ast_node();

  /**
   * \brief checks if this object is an array
   */
  bool is_array();

  /**
   * \brief checks if this object is a reference
   */
  bool is_reference();

  /**
   * \brief checks if this object is a pointer
   */
  bool is_pointer();

  /**
   * \brief  Get function for the name of the module instance variable
   * \return Name as std::string
   */
  virtual std::string get_name();

  /**
   * \fn       get_symbol
   * \brief    return the symbol of the object
   */
  SgVariableSymbol* get_symbol();


  unsigned int number_of_array_elements_;

private:

  /**
   * \brief Disabled copy constructor
   */
  Object(const Object &o);
};

} // end of namespace risc


#endif /* OBJECT_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
