#ifndef STATIC_PATH_INSTANCE_MAPPER_H_INCLUDED_
#define STATIC_PATH_INSTANCE_MAPPER_H_INCLUDED_

#include<vector>
#include<utility>

#include "instance_tree.h"
#include "path_instance_mapper.h"

namespace risc {

class Design;

/**
 * \class PathInstanceMapper
 * \brief This class can be used as a lookup table between an instance id of
 *        a module or hierarchical channel and the instance tree
 */
class StaticPathInstanceMapper: public PathInstanceMapper {

public:

  /**
   * \fn       StaticPathInstanceMapper(InstanceTree root)
   * \brief    This is the default constructor and need a design and a segment
   *           graph. The lookup table will be automatically created.
   */
  StaticPathInstanceMapper(InstanceTree root);

  /**
   * \fn       determine_lookup_table(InstanceTree root)
   * \brief    This function traverses the design and pushes all modules and
   *           hierarchical channel instances into the path_to_id_mapping_
   *           lookup table. The traversing is dependent on the declaration
   *           order of the member variables. Also, it follows depth first
   *           search.
   *           CAUTION: The function Design::reset_instance_counter() must be
   *                    used first! Otherwise wrong intance id's a generated!
   */
  void determine_lookup_table(InstanceTree root);

  /**
   * \fn       get_id(InstanceTree tree)
   * \brief    This function returns for a given tree the corresponding instance
   *           id.
   */
  virtual int get_id(InstanceTree tree);

  /**
   * \fn       get_instance_tree(int id, std::list<InstanceTree> potential_instances)
   * \brief    This function returns for a given id and potential mapping
   *           candidates the corresponding instance tree.
   */
  InstanceTree
  get_instance_tree(int id, std::list<InstanceTree> potential_instances);

  /**
   * \fn       perform_dynamic_analysis
   * \brief    This functions returns true if the path instance mapper should
   *           perform a dynamic analysis
   */
  virtual bool perform_dynamic_analysis();
};

}; // end of namespace risc


#endif /* STATIC_PATH_INSTANCE_MAPPER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
