#ifndef TARGETSOCKET_H_INCLUDED_
#define TARGETSOCKET_H_INCLUDED_

#include <vector>

#include "class.h"
#include "socket.h"

#include "rose.h"

namespace risc {

class TargetSocket: public Socket {

public:

  TargetSocket();
  /**
   * \brief default constructor
   */
  explicit TargetSocket(SgVariableDefinition *ast_node, Class *parent);

  /**
   * \brief gets all the sockets that are possibly connected 
   *		to this socket
   */
  void get_mapped_sockets(std::set<risc::Socket*> &mapped_sockets);

  /**
   * \brief register the b_transport method for this socket
   */
  void register_b_transport(SgFunctionDeclaration*);

  /**
   * \brief get the registered b_transport method
   */
  SgFunctionDeclaration* get_registered_bt();

  /**
   * \brief register the get_direct_mem_ptr method for this socket
   */
  void register_get_dmi(SgFunctionDeclaration* func_decl);

  /**
   * \brief get the registered get_direct_mem_ptr method
   */
  SgFunctionDeclaration* get_registered_get_dmi();

  /**
   * \brief register the nb_transport_fw method for this socket
   */
  void register_nb_transport_fw(SgFunctionDeclaration* func_decl);

  /**
   * \brief get the registered nb_transport_fw method
   */
  SgFunctionDeclaration* get_registered_nb_transport_fw();


private:
  /**
   * \brief Disabled copy constructor
   */
  TargetSocket(const TargetSocket &ip);

  /**
   * \brief registered method for b_transport
   */
  SgFunctionDeclaration* registered_b_transport_;

  /**
   * \brief registered method for get_direct_mem_ptr
   */
  SgFunctionDeclaration* registered_get_dmi;

  /**
   * \brief registered method for nb_transport_fw_
   */
  SgFunctionDeclaration* registered_nb_transport_fw_;
};

typedef std::vector<TargetSocket*> TargetSocketVector;
typedef std::vector<TargetSocket*>::iterator TargetSocketVectorIter;
typedef std::vector<TargetSocket*>::const_iterator TargetSocketVectorConstIter;

}; // end namespace risc

#endif /* TARGETSOCKET_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
