#ifndef REFERENCE_ATTRIBUTES_H_INCLUDED_
#define REFERENCE_ATTRIBUTES_H_INCLUDED_

#include "rose.h"

#include "port_call_path.h"
#include <set>
#include <map>

namespace risc {

namespace sg {

class FunctionParameterReferenceAttribute: public AstAttribute {

public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief    The actual variables that the reference function
   *           parameter really refers to
   */
  std::set<SgVariableSymbol*> mapped_variables_;

  /**
   * \brief    When the parameter is reached by this PortCallPath,
   *           then it should refer to the SgVariableSymbol is the set
   */
  std::map<PortCallPath, std::set<SgVariableSymbol*>> mapped_variables_with_pcp_;
};

class LocalReferenceAttribute: public AstAttribute {
public:

  // for the new rose release
  virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  {
    return CONTAINER_OWNERSHIP;
  }

  /**
   * \brief    The actual variable that the reference function
   *           parameter really refers to
   */
  SgVariableSymbol *mapped_variable_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* REFERENCE_ATTRIBUTES_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
