#include "graph_algorithms.h"

std::set<SgFunctionSymbol*>
risc::sa::ga::dfs(SgFunctionSymbol *root_func_symbol,
    SgIncidenceDirectedGraph &graph)
{
  std::set<SgGraphNode*> node_set = graph.computeNodeSet();

  ColoredNodeVector nodes(node_set.size());
  ColoredNodeVector::iterator root_node;

  for(std::set<SgGraphNode*>::iterator
      iter = node_set.begin();
      iter != node_set.end();
      iter++) {

    ColoredGraphNode temp(*iter,White);
    nodes.push_back(temp);

    SgFunctionDeclaration *func_decl =
      isSgFunctionDeclaration((*iter)->get_SgNode());

    if(root_func_symbol == func_decl->search_for_symbol_from_symbol_table()) {
      root_node = --(nodes.end());
    }
  }

  dfs_visit(*root_node, nodes, graph);

  std::set<SgFunctionSymbol*> called_functions;
  for(ColoredNodeVector::iterator
      iter_node  = nodes.begin();
      iter_node != nodes.end();
      iter_node++) {

    if((*iter_node).second == Black) {
      SgFunctionDeclaration *func_decl =
        isSgFunctionDeclaration((*iter_node).first->get_SgNode());

      called_functions.insert(
          isSgFunctionSymbol(func_decl->search_for_symbol_from_symbol_table()));
    }
  }

  return called_functions;
}

void
risc::sa::ga::dfs_visit(ColoredGraphNode &root, ColoredNodeVector &nodes,
    SgIncidenceDirectedGraph &graph)
{
  root.second = Gray;

  std::set<SgDirectedGraphEdge*> outgoing_edge_set =
    graph.computeEdgeSetOut(root.first);

  for(std::set<SgDirectedGraphEdge*>::iterator
      edge_iter  = outgoing_edge_set.begin();
      edge_iter != outgoing_edge_set.end();
      edge_iter++) {

    //ColoredGraphNode node;
    for(std::vector<ColoredGraphNode>::iterator
        nodes_iter  = nodes.begin();
        nodes_iter != nodes.end(); nodes_iter++) {

      //std::cout << (*edge_iter) << std::endl;
      if((*nodes_iter).first == (*edge_iter)->get_to()) {

        if((*nodes_iter).second == White) {
          dfs_visit(*nodes_iter, nodes, graph);
        }
      }
    }
  }

  root.second = Black;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
