// bug20190404.cc: (WAS: play_TLM2_dmi.cpp)

#include <stdio.h>
#include "systemc.h"
#include "tlm.h"
#include "tlm_utils/simple_initiator_socket.h"
#include "tlm_utils/simple_target_socket.h"
#include <string>
using namespace sc_core;
using namespace std;

#define VIDEO_DELAY     33330, SC_US
#define HALF_VIDEO_DELAY 33330/2, SC_US
#define CLIP_LENGTH     1, SC_SEC
#define VIDEOLOAD       10000000
#define MEMORY_SIZE     0x10000
#define ADDRESS_VIDEO   0x00000

int IntentionalConflict = 0;	// this shall show in the conflict table

SC_MODULE(Memory4p)
{
  public:
    tlm_utils::simple_target_socket<Memory4p>  s0;
    tlm_utils::simple_target_socket<Memory4p>  s1;

    SC_CTOR(Memory4p)
    {
      M = new unsigned char[MEMORY_SIZE];
      s0.register_b_transport(this, &Memory4p::access);
      s1.register_b_transport(this, &Memory4p::access);
      s0.register_get_direct_mem_ptr(this, &Memory4p::get_direct_mem_ptr);
      s1.register_get_direct_mem_ptr(this, &Memory4p::get_direct_mem_ptr);
    }

    void access(tlm::tlm_generic_payload& trans, sc_time& delay)
    {
      tlm::tlm_command cmd = trans.get_command();
      sc_dt::uint64    adr = trans.get_address();
      unsigned char   *ptr = trans.get_data_ptr();
      unsigned int     len = trans.get_data_length();
      unsigned int     wid = trans.get_streaming_width();
      unsigned char   *byt = trans.get_byte_enable_ptr();
      if (  (adr + len <= MEMORY_SIZE)
          &&(ptr != 0)
          &&(wid == len)
          &&(byt == 0))
      {
        if (cmd == tlm::TLM_READ_COMMAND)
        {
          memcpy(ptr, &M[adr], len);
	  IntentionalConflict++;
          trans.set_response_status(tlm::TLM_OK_RESPONSE);
        }
        else if (cmd == tlm::TLM_WRITE_COMMAND)
        {
          memcpy(&M[adr], ptr, len);
	  IntentionalConflict++;
          trans.set_response_status(tlm::TLM_OK_RESPONSE);
        }
        else
        {
          trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
          return;
        }
      }
      else
      {
        trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
        return;
      }
    }

    bool get_direct_mem_ptr(tlm::tlm_generic_payload& trans,
				tlm::tlm_dmi &dmi_data)
    {
      tlm::tlm_command cmd = trans.get_command();
      sc_dt::uint64    adr = trans.get_address();
      if (!(  (adr < MEMORY_SIZE)
	    &&(  (cmd == tlm::TLM_READ_COMMAND)
	       ||(cmd == tlm::TLM_WRITE_COMMAND))))
      {
        trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
        return false;
      }
      dmi_data.allow_read_write();
      dmi_data.set_dmi_ptr(M);
      dmi_data.set_start_address(0);
      dmi_data.set_end_address(MEMORY_SIZE-1);
      dmi_data.set_read_latency(SC_ZERO_TIME);
      dmi_data.set_write_latency(SC_ZERO_TIME);
      IntentionalConflict++;
      return true;
    }

  private:
    unsigned char *M;
};

SC_MODULE(Stimulus)
{
  tlm_utils::simple_initiator_socket<Stimulus> MultiMediaStream;

  sc_event &VCodecStart, &VCodecDone;

  long long video;

  bool testRD_noWait(void)  // RD: testing code decomposition
  {
    return (42 == 27+15);
  }
  void testZC_withWait(void)	// ZC: testing code decomposition
  {
    1 == 0 + 1;
    wait(SC_ZERO_TIME);
    4 == 2 + 2;
  }
  bool testRD_withWait(void)	// RD: testing code decomposition
  {
    bool result = (42 == 27+15);
    wait(SC_ZERO_TIME);
    return (result);
  }

  void Video()
  {
    sc_time ClipLength(CLIP_LENGTH);
    sc_time VideoDelay(VIDEO_DELAY);
    sc_time t1(0, SC_US);
    tlm::tlm_generic_payload trans0;
    tlm::tlm_dmi dmi0;
    unsigned char *ptr = NULL;

    testZC_withWait(), 1; // ZC: testing code decomposition

    testRD_noWait();	// RD: testing code decomposition
    testRD_withWait();	// RD: testing code decomposition

    testRD_noWait(),  testRD_noWait();	// RD: testing code decomposition
    testRD_withWait(),testRD_withWait();// RD: testing code decomposition

    if (testRD_noWait())	// RD: testing code decomposition
    {
      video += 111;
    }
    if (testRD_withWait())	// RD: testing code decomposition
    {
      video += 222;
    }

    if ((11+22==33) && testRD_noWait())	// RD: testing code decomposition
    {
      video += 333;
    }
    if ((11+22==33) && testRD_withWait())	// RD: testing code decomposition
    {
      video += 444;
    }

    (11+22==33) && testRD_withWait();	// ZC: testing code decomposition

    trans0.set_command(tlm::TLM_WRITE_COMMAND);
    trans0.set_address(ADDRESS_VIDEO);
    trans0.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    if (  (!MultiMediaStream->get_direct_mem_ptr(trans0, dmi0)) // BUG: socket call code not in SG!
          ||(!dmi0.is_write_allowed())
          ||(dmi0.get_start_address() > ADDRESS_VIDEO)
          ||(dmi0.get_end_address() < ADDRESS_VIDEO+sizeof(video)-1)
          ||(!(ptr = dmi0.get_dmi_ptr())))
    {
      SC_REPORT_FATAL(name(), "DMI write request for video failed!");
    }
    while(t1 < ClipLength)
    {
      t1 += VideoDelay;
      memcpy(&ptr[ADDRESS_VIDEO], &video, sizeof(video));
      VCodecStart.notify(SC_ZERO_TIME);
      wait(VCodecDone);
    }
  }

  SC_HAS_PROCESS(Stimulus);

  Stimulus(sc_module_name n,
           sc_event &Vstart, sc_event &Vdone)
  : sc_module(n)
  , VCodecStart(Vstart), VCodecDone(Vdone)
  {
    video = 3;
    SC_THREAD(Video);
  }
};

SC_MODULE(VideoCodec)
{
  tlm_utils::simple_initiator_socket<VideoCodec> s1;
  tlm_utils::simple_initiator_socket<VideoCodec> s2;

  sc_event &VCodecStart, &VCodecDone;

  long long d, outFrame, inFrame;
  
  long long decode(long long x)
  {
    int i;

    for (i = 0; i < VIDEOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    tlm::tlm_generic_payload trans0;
    tlm::tlm_dmi dmi0;
    unsigned char *ptr = NULL;

    trans0.set_command(tlm::TLM_READ_COMMAND);
    trans0.set_address(ADDRESS_VIDEO);
    trans0.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    if (  (!s1->get_direct_mem_ptr(trans0, dmi0))
          ||(!dmi0.is_read_allowed())
          ||(dmi0.get_start_address() > ADDRESS_VIDEO)
          ||(dmi0.get_end_address() < ADDRESS_VIDEO+sizeof(inFrame)-1)
          ||(!(ptr = dmi0.get_dmi_ptr())))
    {
      SC_REPORT_FATAL(name(), "DMI read request for video failed!");
    }
    d = 1;

    while(1)
    {
      wait(VCodecStart);

      memcpy(&inFrame, &ptr[ADDRESS_VIDEO], sizeof(inFrame));

      wait(HALF_VIDEO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_VIDEO_DELAY);

      sc_time delay1 = SC_ZERO_TIME;
      tlm::tlm_generic_payload trans1; 
      trans1.set_command(tlm::TLM_WRITE_COMMAND);
      trans1.set_address(0);
      trans1.set_data_ptr(reinterpret_cast<unsigned char*>(&outFrame));
      trans1.set_data_length(sizeof(outFrame));
      trans1.set_streaming_width(sizeof(outFrame));
      trans1.set_byte_enable_ptr(0);
      trans1.set_dmi_allowed(false);
      trans1.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
      //s2->b_transport(trans1, delay1); // original code, OK
      1,s1->b_transport(trans1, delay1); // socket call not followed into SG! (BUG version 2)
      if (trans1.get_response_status() != tlm::TLM_OK_RESPONSE)
      {
        SC_REPORT_FATAL(name(), trans1.get_response_string().c_str());
      }

      VCodecDone.notify(SC_ZERO_TIME);
    }
  }

  SC_HAS_PROCESS(VideoCodec);

  VideoCodec(sc_module_name n,
           sc_event &Vstart, sc_event &Vdone)
  : sc_module(n)
  , VCodecStart(Vstart), VCodecDone(Vdone)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(DUT)
{
  tlm::tlm_target_socket< > MultiMediaStream;
  tlm::tlm_initiator_socket< > VideoOut;

  sc_event &VCodecStart, &VCodecDone;

  Memory4p   memory;
  VideoCodec video;

  DUT(sc_module_name n,
           sc_event &Vstart, sc_event &Vdone)
  : sc_module(n)
  , VCodecStart(Vstart), VCodecDone(Vdone)
  , memory("memory")
  , video("video", VCodecStart, VCodecDone)
  {
    MultiMediaStream.bind(memory.s0);
    video.s1.bind(memory.s1);
    video.s2.bind(VideoOut);
  }
};

SC_MODULE(Display)
{
  tlm_utils::simple_target_socket<Display> VideoIn;

  long long Frame;

  void display(long long )
  {
  }

  void receive(tlm::tlm_generic_payload& trans, sc_time& delay)
  {
    tlm::tlm_command cmd = trans.get_command();
    sc_dt::uint64    adr = trans.get_address();
    unsigned char   *ptr = trans.get_data_ptr();
    unsigned int     len = trans.get_data_length();
    unsigned int     wid = trans.get_streaming_width();
    unsigned char   *byt = trans.get_byte_enable_ptr();
    if (  (cmd == tlm::TLM_WRITE_COMMAND)
        &&(adr == 0)
        &&(ptr != 0)
        &&(len == sizeof(Frame))
        &&(wid == sizeof(Frame))
        &&(byt == 0))
    {
      memcpy(&Frame, ptr, len);
      trans.set_response_status(tlm::TLM_OK_RESPONSE);
    }
    else
    {
      trans.set_response_status(tlm::TLM_GENERIC_ERROR_RESPONSE);
      return;
    }
    display(Frame);
  }

  SC_CTOR(Display)
  {
    VideoIn.register_b_transport(this, &Display::receive);
  }
};

SC_MODULE(Monitor)
{
  tlm::tlm_target_socket< > VideoIn;
  Display screen;

  SC_CTOR(Monitor):
        screen("screen")
  {
    VideoIn.bind(screen.VideoIn);
  }
};

SC_MODULE(Top)
{
  sc_event VCodecStart, VCodecDone;

  Stimulus       stimulus;
  DUT            dut;
  Monitor        monitor;

  SC_CTOR(Top)
  : stimulus("stimulus",
        VCodecStart, VCodecDone)
  , dut("dut",
        VCodecStart, VCodecDone)
  , monitor("monitor")
  {
    stimulus.MultiMediaStream.bind(dut.MultiMediaStream);
    dut.VideoOut.bind(monitor.VideoIn);
  }
};

int sc_main(int argc, char* argv[])
{
  Top top("top");

  printf("Playing...\n");
  sc_start();
  printf("Done!\n");
  return 0;
}

// EOF
