// one_notify_one_wait2.cc
// 04/18/18 RD   adjusted Rnd() to use reference instead of pointer
// 04/11/17 RD   version with random workload
// 04/11/17 RD   added assertions
// 04/04/17 ZC   initial version

#include <stdio.h>
#include "systemc.h"

#define ITERATIONS 500
#define WORKLOAD   1000000

// deterministic MT-safe pseudo random number in range [0...Max]
// (based on example by 'man rand', 4/8/17, RD)
int Rnd(unsigned long &Next, int Max)
{
    Next = Next * 1103515245 + 12345;
    return((unsigned)(Next/65536) % (Max+1));
}

// simulated workload

unsigned long Work(int Load)
{
    int Count;
    unsigned long Result = 42;

    for(Count=0; Count<Load; Count++)
    {
	Result *= 17;
    }
    return Result;
}

class Main_Block : public sc_module
{
    public:
	sc_event e;
	int s, r;
	unsigned long sw, rw;

	void S()
	{
		unsigned long RndNext = 1;

		printf("S starting\n");
		s = 1;
		wait(1,SC_NS);
		for(int i=0;i<ITERATIONS;i++){
			sw = Work(Rnd(RndNext, WORKLOAD));
			e.notify(SC_ZERO_TIME);
			sw = Work(Rnd(RndNext, WORKLOAD));
			wait(2,SC_NS);
			sw = Work(Rnd(RndNext, WORKLOAD));
			s++;
		}
		assert(s == ITERATIONS+1);
		printf("S done\n");
	}

	void R()
	{
		unsigned long RndNext = 2;

		printf("R starting\n");
		r = 1;
		for(int i=0;i<ITERATIONS;i++){
			rw = Work(Rnd(RndNext, WORKLOAD));
			wait(e);
			rw = Work(Rnd(RndNext, WORKLOAD));
			r++;
		}
		printf("R done\n");
		assert(r == ITERATIONS+1);
	}

	Main_Block(sc_module_name name)
	: sc_module(name)
	{
		s = r = 0;
		SC_THREAD(S);
		SC_THREAD(R);
	}
	SC_HAS_PROCESS(Main_Block);
};

class Platform : public sc_module
{
    public:
	Main_Block m;

	Platform(sc_module_name name)
	: sc_module(name),
	m("m")
	{
		printf("one_notify_one_wait2 constructor\n");
	}
	SC_HAS_PROCESS(Platform);
};

int sc_main(int argc, char* argv[])
{
    Platform p("p");

    printf("one_notify_one_wait2 starting\n");
    assert(p.m.s == 0);
    assert(p.m.r == 0);
    sc_start();
    assert(p.m.s == ITERATIONS+1);
    assert(p.m.r == ITERATIONS+1);
    printf("one_notify_one_wait2 done\n");
    return 0;
}

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8 expandtab: */
