// sensitive2.cc:
//
// 07/07/18 RD	regression test for threads with static sensitivity

#include <assert.h>
#include "systemc.h"

#define T 10

SC_MODULE(Top) {
	sc_event tick_even, tick_odd, tick_ten;
	int n_clock;
	int work_even, work_odd, work_ten, work_twice_as_hard, work_tasks;
	sc_time tc;
	sc_time tw1, tw2, tw10, tw2x, twt;

	SC_CTOR(Top) {
		n_clock = 0;
		work_even = work_odd = work_ten = work_twice_as_hard = work_tasks = 0;
		tc = SC_ZERO_TIME;
		tw1 = tw2 = tw10 = tw2x = twt = SC_ZERO_TIME;
		SC_THREAD(clock);
		SC_THREAD(worker_even);
		sensitive << tick_even;
		SC_THREAD(worker_odd);
		sensitive << tick_odd;
		SC_THREAD(worker_ten);
		sensitive << tick_ten;
		SC_THREAD(worker_twice_as_hard);
		sensitive << tick_odd << tick_even;
		sensitive << tick_ten;
		SC_THREAD(worker_tasks);
		sensitive << tick_odd << tick_even;
	}

	void clock() {
		for(int i=0; i<T; i+=1) {
			wait(1,SC_SEC);
			std::cout << sc_time_stamp().to_string() << ": clock tick! ";
			if (i%10==0) {
				tick_ten.notify(SC_ZERO_TIME);
				std::cout << "ten, ";
			}
			if (i%2==0) {
				tick_even.notify(SC_ZERO_TIME);
				std::cout << "even" << std::endl;
			} else {
				tick_odd.notify(SC_ZERO_TIME);
				std::cout << "odd" << std::endl;
			}
			n_clock++;
		}
		tc = sc_time_stamp();
	}

	void worker_even() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked even shift!\n";
			work_even++;
			tw2 = sc_time_stamp();
		}
	}

	void worker_odd() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked odd shift!\n";
			work_odd++;
			tw1 = sc_time_stamp();
		}
	}

	void worker_ten() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked ten shift!\n";
			work_ten++;
			tw10 = sc_time_stamp();
		}
	}

	void worker_twice_as_hard() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked another shift!\n";
			work_twice_as_hard++;
			tw2x = sc_time_stamp();
		}
	}

	void worker_tasks() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 1.\n";
		work_tasks++;
		twt = sc_time_stamp();
		{	wait();
			std::cout << sc_time_stamp().to_string() << ": worked task 2.\n";
			work_tasks++;
			twt = sc_time_stamp();
		}
		for(int i=0; i<3; i++)
		{	wait();
			std::cout << sc_time_stamp().to_string() << ": worked on tasks 3,4,5.\n";
			work_tasks++;
			twt = sc_time_stamp();
		}
		task6();
		task7and8();
	}

	void task6() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 6.\n";
		work_tasks++;
		twt = sc_time_stamp();
	}

	void task7and8() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 7.\n";
		work_tasks++;
		twt = sc_time_stamp();
		task8();
	}

	void task8() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 8.\n";
		work_tasks++;
		twt = sc_time_stamp();
	}
};

Top top("top");

int sc_main(int argc, char** argv) {
	const sc_time tc(T, SC_SEC);
	const sc_time tw1(T, SC_SEC);
	const sc_time tw2(T-1, SC_SEC);
	const sc_time tw10(T/10*10-9, SC_SEC);
	const sc_time tw2x(T, SC_SEC);
	const sc_time twt(8, SC_SEC);

	cout << "sensitive2 test starts" << endl;
	sc_start();
	assert(top.n_clock == T);
	assert(top.work_odd == T/2);
	assert(top.work_even == T/2);
	assert(top.work_ten == T/10);
	assert(top.work_twice_as_hard == T);
	assert(top.work_tasks == 8);
	assert(top.tc == tc);
	assert(top.tw1 == tw1);
	assert(top.tw2 == tw2);
	assert(top.tw10 == tw10);
	assert(top.tw2x == tw2x);
	assert(top.twt == twt);
	cout << "sensitive2 test done." << endl;
	return 0;
}

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8: */
