#include "systemc.h"

class DeviceModel : public sc_module {
public:
SC_HAS_PROCESS(DeviceModel);
DeviceModel(sc_module_name name_) : sc_module(name_), count_(10) {
	SC_THREAD(run_test);
}

void run_test() {
    sc_time loop_time(500,SC_NS); /* 4/5/19, RD */
    while (--count_ >= 0) {
	printf("Looping at time %llu\n",sc_time_stamp().value());
        wait(loop_time);
    }
}

void set_count(int count) {count_ = count; }
int get_count() { return count_; }

private:
int count_;
};

void set_event_posted(bool& event_posted_, uint64& next_posted_event_, sc_time& time_quantum_) {
    if (sc_pending_activity_at_current_time()) {
      	std::cout << "DEBUG: pending activity at current time\n";
        event_posted_ = true;
        next_posted_event_ = sc_time_stamp().value();
        return;
    }
    sc_time next_time;
    if (!sc_get_curr_simcontext()->next_time(next_time)) {
	std::cout << "DEBUG: no next time\n";
        return;
    }
    uint64 next_systemc_event = next_time.value();
    if (event_posted_) {
        if (time_quantum_ != SC_ZERO_TIME) {
            printf("DEBUG: event posted (0)\n");
            return;
        } else if (next_posted_event_ == next_systemc_event) {
            printf("DEBUG: event posted (%llu)\n", next_posted_event_);
            return;
        }
    }
    printf("DEBUG: event posted\n");
    event_posted_ = true;
}

void synch_systemc_with_vp(sc_time& vp_simulation_time_) {
    sc_dt::uint64 time = sc_time_stamp().value();
    uint64 vp_time = vp_simulation_time_.value();
    uint64 time_diff = vp_time - time;
    if (time >= vp_time || sc_core::sc_get_status() == sc_core::SC_STOPPED)
        return;
    if (sc_core::sc_get_status() == sc_core::SC_RUNNING)
        return;
    printf("DEBUG: sc_start(%llu) at %llu\n" , time_diff, time);
    sc_start(sc_time::from_value(time_diff));
    printf("DEBUG: sc_start returned at %llu ps\n", sc_time_stamp().value());
    assert(vp_time == sc_time_stamp().value());
}

void simulate_events_at_current_time() {
    if (sc_core::sc_get_status() == sc_core::SC_RUNNING)
        return;
    printf("DEBUG: simulate events at current time\n");
    unsigned int delta_count = 0;
    while (sc_pending_activity_at_current_time()) {
        sc_start(sc_core::SC_ZERO_TIME);
        ++delta_count;
        assert(delta_count < 999);
    }
    printf("DEBUG: done with events at current time\n");
}

int sc_main(int argc, char* argv[]) {
    DeviceModel t("t");
    bool event_posted_ = false;
    long long unsigned int next_posted_event_;
    sc_time time_quantum_(500,SC_NS);
    sc_time vp_simulation_time_(SC_ZERO_TIME);

    sc_get_curr_simcontext()->initialize(true);
    set_event_posted(event_posted_, next_posted_event_, time_quantum_);
    while(event_posted_ && sc_core::sc_get_status() != sc_core::SC_STOPPED) {
        event_posted_ = false;
        synch_systemc_with_vp(vp_simulation_time_);
        simulate_events_at_current_time();
        set_event_posted(event_posted_,next_posted_event_,time_quantum_);
        vp_simulation_time_ += time_quantum_;
    }
    assert(t.get_count() == -1);
    return 0;
}

