#include "systemc.h"

class ChannelInterface: virtual public sc_interface
{
public:
  virtual void func() = 0;
};

class Channel:
  public sc_channel,
  public ChannelInterface {

public:
  Channel(sc_module_name name):
    sc_channel(name)
  { }

  sc_event event;

  void func()
  {
    event.notify(SC_ZERO_TIME);
    wait(event);
    v++;
  }

  int v;
};

class Ping: public sc_module {
public:

  SC_HAS_PROCESS(Ping);
  sc_port<ChannelInterface> port_i;

  Ping(sc_module_name mn): sc_module(mn)
  {
    SC_THREAD(thread1);
  }

  void thread1()
  {
    i++;
    port_i->func();
  }

  int i;
};

class Pong: public sc_module {
public:

  SC_HAS_PROCESS(Pong);
  sc_port<ChannelInterface> port_o;

  int o;

  void thread1()
  {
    o++;
    port_o->func();
  }

  Pong(sc_module_name mn): sc_module(mn)
  {
    SC_THREAD(thread1);
  }
};

class Top: public sc_module {
public:
  Ping ping1;
  Ping ping2;
  Pong pong2;
  Pong pong1;
  Channel c1;
  Channel c2;
  Channel c3;
  Channel c4;

  Top(sc_module_name mn):
    ping1("ping1"),
    ping2("ping2"),
    pong2("pong2"),
    pong1("pong1"),
    c1("channel1"),
    c2("channel2"),
    c3("channel3"),
    c4("channel4")
  { }

  void before_end_of_elaboration()
  {
    ping1.port_i.bind(c3);
    pong1.port_o.bind(c3);

    ping2.port_i.bind(c4);
    pong2.port_o.bind(c4);
  }
};


int sc_main(int argc, char **argv)
{
  Top t("top");

  sc_start();
  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
